{- Language 1: floating-point arithmetic + holes -}

-- {-# OPTIONS_GHC -fwarn-incomplete-patterns #-}

module Demo1 where

import qualified Demo0

data Exp
  = Con Float
  | Add Exp Exp
  | Mul Exp Exp
  | Pow Exp Exp
  | Todo String

avo = Con 6.022 `Mul`
  (Con 10 `Pow` Todo "Revise chemistry")

{- Evaluate an Exp -}
eval :: Exp -> Float
eval (Con k) = k
eval (Add e1 e2) = eval e1 + eval e2
eval (Mul e1 e2) = eval e1 * eval e2
eval (Pow e1 e2) = eval e1 ** eval e2
-- {- Missing the case below: saved by GHCs completeness-checker (if enabled) -}
-- eval (Todo msg) = error $ unwords
--   ["Hole evaluated!","Message:" , msg]

{- > eval avo -}

{- 💩 crashes on the user's computer! -}

{- 💡 write a todo-checker -}

{- Check that an Exp contains no holes -}
check :: Exp -> Bool
check (Con k)     = True
check (Todo str)  = False
check (Add e1 e2) = check e1 && check e2
check (Mul e1 e2) = check e1 && check e2
check (Pow e1 e2) = check e1 && check e2

-- {- Evaluate an Exp.
--    If there are holes, returns Nothing (before evaluation) -}
-- safeEval :: Exp -> Maybe Float
-- safeEval ast =
--   if (check ast)
--   then Just $ eval ast
--   else Nothing

{- > safeEval avo -}

{- 💩 'Nothing' is not a nice error message -}

{- 💡 collect holes -}

{- Find all the holes in an Exp -}
infer :: Exp -> [String]
infer (Con k)     = []
infer (Todo msg)  = [msg]
infer (Add e1 e2) = infer e1 ++ infer e2
infer (Mul e1 e2) = infer e1 ++ infer e2
infer (Pow e1 e2) = infer e1 ++ infer e2

-- {-  Evaluate an Exp.
--   If there are holes, return them instead. -}
-- safeEval :: Exp -> Either [String] Float
-- safeEval ast =
--   let ts = infer ast in
--   if ts == []
--   then Right (eval ast)
--   else Left ts

{- 💩 eval is messy
  * contains unreachable code (for error handling)
  * repeats Demo0.eval -}

{- 💡 re-use Demo0.eval -}

-- {- Convert an Exp to an Exp₀
--   If there are holes, return them instead -}
-- validate :: Exp -> Either [String] Demo0.Exp
-- validate (Todo msg)  = Left ([msg])
-- validate (Con k)     = Demo0.Con <$> pure k
-- validate (Add e1 e2) = Demo0.Add <$> validate e1 <*> validate e2
-- validate (Mul e1 e2) = Demo0.Mul <$> validate e1 <*> validate e2
-- validate (Pow e1 e2) = Demo0.Pow <$> validate e1 <*> validate e2

-- {- Evaluate an Exp.
--   If there are holes, return them instead. -}
-- safeEval :: Exp -> Either [String] Float
safeEval ast = Demo0.eval <$> validate ast 

{- ✅ safeEval avo -}

{- 🕵 -}

h = Todo "6.6 I think???"  `Mul`
  (Con 10 `Pow` Todo "Revise physics")

{- > safeEval h -}

{- 💩 only the 1st error! -}

{- 💡 change implementation of <*> -}

{- Like Either, but has a <*> that collects err s -}
{- Credit to https://blog.ploeh.dk/2018/11/05/applicative-validation/ -}
data Valid err r = Err err | OK r deriving (Show)

instance Functor (Valid err) where
  fmap f (Err e) = Err e
  fmap f (OK r) = OK (f r)
instance Monoid err => Applicative (Valid err) where
  pure = OK
  (Err x) <*> (Err y) = Err (x <> y)  -- Either drops y here 
  (Err x) <*> (OK _)  = Err x
  (OK _)  <*> (Err y) = Err y
  (OK f)  <*> (OK x)  = OK (f x)
  (Err x) *> (Err y) = Err (x <> y)
  (Err x) *> (OK _)  = Err x
  (OK _) *> y  = y
instance (Monoid err) => Monad (Valid err) where
  (Err x) >>= f = Err x
  (OK x) >>= f = f x
  (>>) = (*>)

{- 🏋️ reimplement Valid as a newtype around Either -}

validate :: Exp -> Valid [String] Demo0.Exp
validate (Todo msg)  = Err ([msg])
validate (Con k)     = Demo0.Con <$> pure k
validate (Add e1 e2) = Demo0.Add <$> validate e1 <*> validate e2
validate (Mul e1 e2) = Demo0.Mul <$> validate e1 <*> validate e2
validate (Pow e1 e2) = Demo0.Pow <$> validate e1 <*> validate e2

safeEval :: Exp -> Valid [String] Float

{-✅ safeEval h -}

{- 🏋️ reimplement using algebras:
  * eval
  * check
  * infer
  * validate
-}
