{- Language 2: floating-point arithmetic + let-bindings -}

module Demo2 where

import Data.Set
import Data.Map
import qualified Data.Set as Set
import qualified Data.Map as Map

data Exp
  = Con Float
  | Add Exp Exp
  | Mul Exp Exp
  | Pow Exp Exp
  | Var String
  | Let String Exp Exp

energy =
  Let "m" (Con 75) (
  Let "g" (Con 9.81) ( 
  Let "h" (Con 10) (
    Var "m" `Mul` (Var "g" `Mul` Var "h")
  )))

eval :: Exp -> Float

-- eval (Con k)    = k
-- eval (Add e1 e2) = eval e1 + eval e2
-- eval (Mul e1 e2) = eval e1 * eval e2
-- eval (Pow e1 e2) = eval e1 ** eval e2
-- eval (Var v) = ??? -- STUCK
-- eval (Let v head body) = ???

eval = go Map.empty where
  go :: Map String Float -> Exp -> Float
  go env (Con k)     = k
  go env (Add e1 e2) = go env e1 +  go env e2
  go env (Mul e1 e2) = go env e1 *  go env e2
  go env (Pow e1 e2) = go env e1 ** go env e2
  go env (Var v) = env ! v
  go env (Let v head body) = go env body where
      head' = go env head
      env = Map.insert v head' env

{- ✅ eval energy -}

{- 🕵 -}

mass =
  (Let "volume" (Var "length" `Mul` (Var "width" `Mul` Var "height"))
  (Let "density" (Con 1.0)
  (Var "volume" `Mul` Var "density")))

{- > eval mass -}

{- 💩 crashes on the user's computer! -}

{- 💡 write a scope-checker -}

{- Check that all variables in an Exp are in scope -}
check :: Exp -> Bool
check = go Set.empty where
  go :: Set String -> Exp -> Bool
  go s (Con _) = True
  go s (Add e1 e2) = go s e1 && go s e2
  go s (Mul e1 e2) = go s e1 && go s e2
  go s (Pow e1 e2) = go s e1 && go s e2
  go s (Var str) = str `elem` s
  go s (Let str head body) = go s head && go s' body where
    s' = Set.insert str s

-- safeEval :: Exp -> Maybe Float
-- safeEval e =
--   if check e
--   then Just $ eval e
--   else Nothing

{- > safeEval mass -}

{- 💩 'Nothing' is not a nice error message -}

{- Infer the out-of-scope (aka "free") variables in an expression -}
infer :: Exp -> Set String
infer = go Set.empty where
  go :: Set String -> Exp -> Set String
  go s (Con _)     = Set.empty
  go s (Add e1 e2) = go s e1 `Set.union` go s e2
  go s (Mul e1 e2) = go s e1 `Set.union` go s e2
  go s (Pow e1 e2) = go s e1 `Set.union` go s e2
  go s (Var str) =
    if str `elem` s
    then Set.empty
    else Set.singleton str
  go s (Let str head body) = go s head `Set.union` go s' body where
    s' = Set.insert str s

safeEval :: Exp -> Either (Set String) Float
safeEval e =
  let freeVars = infer e in
  if freeVars == Set.empty
  then Right $ eval e
  else Left $ freeVars

{- ✅ safeEval mass -}

{- Question: can we write validate? -}

{- Answer: not in Haskell :(
  * What would be validate's type?
    * ❌ Exp -> Valid [String] Demo0.Exp
      * Demo0.Exp can't handle vars
    * ✅ Exp -> Valid [String] Exp_Scoped
      * where Exp_Scoped enforces well-scoped vars
  * How to write Exp_Scoped?
  * To enforce well-scoped vars need...
      * Advanced Haskell type hackery
      * e.g. PHOAS
      * e.g. fiddly GADT
  * Use Agda instead
-}
