{- Language 3: floating-point/boolean expressions -}

module Demo3 where

import Demo1 (Valid(..))

data Exp
  = Con Float
  | Add Exp Exp
  | Mul Exp Exp
  | Pow Exp Exp
  | If Exp Exp Exp
  | Leq Exp Exp
  deriving Show

seven =
  If (Con 3.0 `Leq` Con 4.0)
    (Con 7.0)
    (Con 9.0)

eval :: Exp -> Either Float Bool
eval (Con k)     = Left k
eval (Add e1 e2) = Left (p +  q) where Left p = eval e1; Left q = eval e2
eval (Mul e1 e2) = Left (p *  q) where Left p = eval e1; Left q = eval e2
eval (Pow e1 e2) = Left (p ** q) where Left p = eval e1; Left q = eval e2
eval (If h t e) = if h' then eval t else eval t where
  Right h' = eval h 
eval (Leq e1 e2) = Right (p <= q) where Left p = eval e1; Left q = eval e2

{- ✅ eval seven -}

{- 💩 eval is ugly -}

{- 🕵 -}

nseven = 
  If (Con 3.0)
    (Con 7.0)
    (Con 9.0)

{- > eval nseven -}

{- 💩 crashes on the user's computer! -}

{- 💡 write type-checker -}

data Type = Float | Bool
  deriving (Show, Eq)

-- check :: Type -> Exp -> Bool
-- check Float (Con _) = True
-- check Float (Add e1 e2) = check Float e1 && check Float e2
-- check Float (Mul e1 e2) = check Float e1 && check Float e2
-- check Float (Pow e1 e2) = check Float e1 && check Float e2
-- check ty (If h t e) = check Bool h && check ty t && check ty e
-- check _ _ = False

-- safeEval :: Type -> Exp -> Maybe (Either Float Bool) 
-- safeEval t e =
--   if check t e
--   then Just $ eval e
--   else Nothing

{- > safeEval Float nseven -}

{- 💩 safeEval should not need a Type param -}
{- 💩 'Nothing' is not a nice error message -}
{- 💩 eval is ugly -}

{- 💡 get rid of the Type param by inferring type -}

-- infer :: Exp -> Maybe Type
-- infer e@(Con _)   = if check Float e then Just Float else Nothing
-- infer e@(Add _ _) = if check Float e then Just Float else Nothing
-- infer e@(Mul _ _) = if check Float e then Just Float else Nothing
-- infer e@(Pow _ _) = if check Float e then Just Float else Nothing
-- infer e@(Leq _ _) = if check Bool e  then Just Bool  else Nothing
-- infer (If h t e) = do
--   t_ty <- infer t
--   e_ty <- infer e
--   if t_ty == e_ty && check Bool h
--   then Just t_ty
--   else Nothing

{- note the re-use of check (to avoid duplication) -}

-- safeEval :: Exp -> Maybe (Either Float Bool) 
-- safeEval e = infer e >> Just (eval e)

{- > safeEval nseven -}

{- 💩 infer, check ugly -}
{- 💩 'Nothing' is not a nice error message -}

{- 💡 use Maybe () instead of Bool to unlock <*> etc. -}

-- check :: Type -> Exp -> Maybe ()
-- check Float (Con _)     = Just ()
-- check Float (Add e1 e2) = () <$ check Float e1 <* check Float e2
-- check Float (Mul e1 e2) = () <$ check Float e1 <* check Float e2
-- check Float (Pow e1 e2) = () <$ check Float e1 <* check Float e2
-- check ty (If h t e)     = () <$ check Bool h <* check ty t <* check ty e
-- check _ _ = Nothing

-- infer :: Exp -> Maybe Type
-- infer e@(Con _)   = Float <$ check Float e
-- infer e@(Add _ _) = Float <$ check Float e
-- infer e@(Mul _ _) = Float <$ check Float e
-- infer e@(Pow _ _) = Float <$ check Float e
-- infer e@(Leq _ _) = Bool  <$ check Bool e
-- infer (If h t e) = do
--   check Bool h
--   t_ty <- infer t
--   t_ty <$ check t_ty e

-- safeEval :: Exp -> Maybe (Either Float Bool) 
-- safeEval e = infer e >> Just (eval e)

{- Unchanged: safeEval nseven -}

{- 💩 'Nothing' is not a nice error message -}

{- 💡 replace Maybe with Either -}

data CheckingErr = Expected Type {- in -} Exp
  deriving Show

-- check :: Type -> Exp -> Either CheckingErr ()
-- check Float (Con _)     = Right ()
-- check Float (Add e1 e2) = () <$ check Float e1 <* check Float e2          -- unchanged
-- check Float (Mul e1 e2) = () <$ check Float e1 <* check Float e2          -- unchanged
-- check Float (Pow e1 e2) = () <$ check Float e1 <* check Float e2          -- unchanged
-- check ty (If h t e)     = () <$ check Bool h <* check ty t <* check ty e  -- unchanged
-- check expected_ty e     = Left $ Expected expected_ty e

-- infer :: Exp -> Either CheckingErr Type
-- infer e@(Con _)   = Float <$ check Float e
-- infer e@(Add _ _) = Float <$ check Float e  -- unchanged
-- infer e@(Mul _ _) = Float <$ check Float e  -- unchanged
-- infer e@(Pow _ _) = Float <$ check Float e  -- unchanged
-- infer e@(Leq _ _) = Bool  <$ check Bool  e  -- unchanged
-- infer (If h t e) = do                       -- unchanged
--   check Bool h                              -- unchanged
--   t_ty <- infer t                           -- unchanged
--   t_ty <$ check t_ty e                      -- unchanged

-- safeEval :: Exp -> Either CheckingErr (Either Float Bool) 
-- safeEval e = infer e >> Right (eval e)

-- {- ✅ safeEval nseven -}

neight = 
  If (Con 3.0)
    (If (Con 0.0) (Con 0.0) (Con 0.0))
    (Con 8.0)

{- > safeEval neight -}

{- 💩 only one error reported! -}

{- 💡 bring back Valid -}

check :: Type -> Exp -> Valid [CheckingErr] ()
check Float (Con _)     = OK ()
check Float (Add e1 e2) = () <$ check Float e1 <* check Float e2          -- unchanged
check Float (Mul e1 e2) = () <$ check Float e1 <* check Float e2          -- unchanged
check Float (Pow e1 e2) = () <$ check Float e1 <* check Float e2          -- unchanged
check ty (If h t e)     = () <$ check Bool h <* check ty t <* check ty e  -- unchanged
check expected_ty e     = Err [Expected expected_ty e]

infer :: Exp -> Valid [CheckingErr] Type
infer e@(Con _)   = Float <$ check Float e  -- unchanged
infer e@(Add _ _) = Float <$ check Float e  -- unchanged
infer e@(Mul _ _) = Float <$ check Float e  -- unchanged
infer e@(Pow _ _) = Float <$ check Float e  -- unchanged
infer e@(Leq _ _) = Bool  <$ check Bool  e  -- unchanged
infer (If h t e) = do                       -- unchanged
  check Bool h                              -- unchanged
  t_ty <- infer t                           -- unchanged
  t_ty <$ check t_ty e                      -- unchanged

safeEval :: Exp -> Valid [CheckingErr] (Either Float Bool) 
safeEval e = infer e >> OK (eval e)

{- ✅ safeEval neight -}

{- 💩 eval is still ugly -}

{- 🏋🏋 implement...
  * A type Exp_Float allowing *only* Float-typed Exps
  * validate  :: Exp -> Valid [CheckingErr] Exp_Float
  * evalFloat :: Exp_Float -> Float
  * safeEval in terms of `validate` and `evalFloat`

Hint: you will also need to implement
  corresponding defintions for Bool-typed Exp s
  (or use a GADT)
-}
